# -*- coding: utf-8 -*-

from .tree import (
    get_node, get_nodes, get_root,
    create_node, create_nodes,
    get_language, set_language,
    update_node_value, maybe_update_node_values,
    update_node_children, maybe_update_node_children,
    )
from .entities import update_entities
from .attributes import merge_l10n_attributes, refactor_relationship
from .properties import delete_unused_properties, merge_properties

__all__ = [
    'get_node', 'get_nodes', 'get_root',
    'create_node', 'create_nodes',
    'get_language', 'set_language',
    'update_node_value', 'maybe_update_node_values',
    'update_node_children', 'maybe_update_node_children',

    'update_entities',
    'merge_l10n_attributes',
    'refactor_relationship',
    'delete_unused_properties', 'merge_properties',

    '__copyright__', '__license__',
    ]
__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
